"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.executeTransaction = exports.IsolationLevel = exports.awaitWithTimeout = exports.sleep = void 0;
const error_1 = require("./error");
/**
 * Sleep for a given amount of milliseconds
 * @param milliseconds The time in milliseconds to sleep
 * @returns The (void) promise to await
 */
const sleep = (milliseconds) => __awaiter(void 0, void 0, void 0, function* () { return new Promise((resolve) => setTimeout(resolve, milliseconds)); });
exports.sleep = sleep;
/**
 * Run a promise but make sure to wait only a maximum amount of time for it to finish.
 * @param promise The promise to execute
 * @param timeoutMs The amount of time in milliseconds to wait for the promise to finish
 * @param failureMessage The message for the error if the timeout was reached
 * @returns The promise return value or a timeout error is thrown
 */
const awaitWithTimeout = (promise, timeoutMs, failureMessage) => {
    let timeoutHandle;
    const timeoutPromise = new Promise((_resolve, reject) => {
        timeoutHandle = setTimeout(() => reject(new Error(failureMessage)), timeoutMs);
    });
    return Promise.race([promise(), timeoutPromise]).then((result) => {
        clearTimeout(timeoutHandle);
        return result;
    });
};
exports.awaitWithTimeout = awaitWithTimeout;
/**
 * PostgreSQL available isolation levels. The isolation level "Read uncommitted"
 * is the same as "Read committed" in PostgreSQL. And the readonly variants are
 * not usable as the inbox message must be marked as processed or the
 * "finished_attempts" counter is updated for the message.
 */
var IsolationLevel;
(function (IsolationLevel) {
    IsolationLevel["Serializable"] = "SERIALIZABLE";
    IsolationLevel["RepeatableRead"] = "REPEATABLE READ";
    IsolationLevel["ReadCommitted"] = "READ COMMITTED";
})(IsolationLevel || (exports.IsolationLevel = IsolationLevel = {}));
/**
 * Open a transaction and execute the callback as part of the transaction.
 * @param pool The PostgreSQL database pool
 * @param callback The callback to execute DB commands with.
 * @param isolationLevel The database transaction isolation level. Falls back to the default PostgreSQL transaction level if not provided.
 * @param logger A logger instance for logging trace up to error logs on DB based client errors
 * @returns The result of the callback (if any).
 * @throws Any error from the database or the callback.
 */
const executeTransaction = (pool, callback, isolationLevel, logger) => __awaiter(void 0, void 0, void 0, function* () {
    let client = undefined;
    const isolation = Object.values(IsolationLevel).includes(isolationLevel)
        ? isolationLevel
        : undefined;
    try {
        client = yield getClient(pool, logger);
        yield client.query(isolation ? `START TRANSACTION ISOLATION LEVEL ${isolation}` : 'BEGIN');
        const result = yield callback(client);
        yield client.query('COMMIT');
        client.release();
        return result;
    }
    catch (err) {
        const error = (0, error_1.ensureError)(err);
        try {
            yield (client === null || client === void 0 ? void 0 : client.query('ROLLBACK'));
            client === null || client === void 0 ? void 0 : client.release(true);
        }
        catch (rollbackError) {
            // We report the initial error - this one is about DB connection issues
        }
        throw error;
    }
});
exports.executeTransaction = executeTransaction;
const getClient = (pool, logger) => __awaiter(void 0, void 0, void 0, function* () {
    const client = yield pool.connect();
    // The pool can return a new or an old client - we must register the event listener but should do so only once
    if (!client.listeners('error').length && logger) {
        client.on('error', (err) => {
            logger.error(err, 'PostgreSQL client error');
        });
    }
    return client;
});
//# sourceMappingURL=utils.js.map