import { ClientBase, ClientConfig } from 'pg';
import { ErrorType } from '../common/error';
import { TransactionalLogger } from '../common/logger';
import { InboxMessage } from '../common/message';
import { TransactionalOutboxInboxConfig } from '../replication/config';
import { TransactionalStrategies } from '../replication/logical-replication-listener';
import { MessageProcessingTransactionLevelStrategy } from '../strategies/message-processing-transaction-level-strategy';
import { MessageRetryStrategy } from '../strategies/message-retry-strategy';
import { PoisonousMessageRetryStrategy } from '../strategies/poisonous-message-retry-strategy';
/** The inbox listener configuration */
export type InboxConfig = TransactionalOutboxInboxConfig & {
    /**
     * Database connection details. The user needs update permission to the inbox.
     */
    pgConfig: ClientConfig;
    settings: TransactionalOutboxInboxConfig['settings'] & {
        /**
         * The maximum number of attempts to handle an incoming inbox message.
         * Defaults to 5 which means a message is handled once initially and up to
         * four more times for retries.
         */
        maxAttempts?: number;
        /**
         * Defines the maximum number of times a message is going to be processed
         * when it is (likely) a poisonous message which is causing a server crash.
         * Defaults to three.
         */
        maxPoisonousAttempts?: number;
    };
};
/**
 * Message handler for a specific aggregate type and message type.
 */
export interface InboxMessageHandler {
    /** The aggregate root type */
    aggregateType: string;
    /** The name of the message created for the aggregate type. */
    messageType: string;
    /**
     * Custom business logic to handle a message that was stored in the inbox. It
     * is fine to throw an error if the message cannot be processed.
     * @param message The inbox message with the payload to handle.
     * @param client The database client that is part of a transaction to safely handle the inbox message.
     * @throws If something failed and the inbox message should NOT be acknowledged - throw an error.
     */
    handle: (message: InboxMessage, client: ClientBase) => Promise<void>;
    /**
     * Custom (optional) business logic to handle an error that was caused by the
     * "handle" method. Please ensure that this function does not throw an error
     * as the finished_attempts counter is increased in the same transaction.
     * @param error The error that was thrown in the handle method.
     * @param message The inbox message with the payload that was attempted to be handled.
     * @param client The database client that is part of a (new) transaction to safely handle the error.
     * @param attempts The number of times that message was already attempted (including the current attempt). And the maximum number of times the message will be attempted.
     * @returns A flag that defines if the message should be retried ('transient_error') or not ('permanent_error')
     */
    handleError?: (error: Error, message: InboxMessage, client: ClientBase, attempts: {
        current: number;
        max: number;
    }) => Promise<ErrorType>;
}
/** Strategies to provide the desired logic for handling specific scenarios */
export interface InboxStrategies extends TransactionalStrategies {
    /**
     * Defines the PostgreSQL transaction level to use for handling an incoming
     * inbox message.
     */
    messageProcessingTransactionLevelStrategy: MessageProcessingTransactionLevelStrategy;
    /**
     * Decides if a message should be retried or not based on the current amount
     * of attempts.
     */
    messageRetryStrategy: MessageRetryStrategy;
    /**
     * Decide based on the message, the poisonous attempts counter (which is
     * already increased by one), and the processing attempts if the message
     * should be retried again or not.
     */
    poisonousMessageRetryStrategy: PoisonousMessageRetryStrategy;
}
/**
 * Initialize the listener to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages.
 * @param logger A logger instance for logging trace up to error logs
 * @param strategies Strategies to provide custom logic for handling specific scenarios
 * @returns Functions for a clean shutdown.
 */
export declare const initializeInboxListener: (config: InboxConfig, messageHandlers: InboxMessageHandler[], logger: TransactionalLogger, strategies?: Partial<InboxStrategies>) => [shutdown: () => Promise<void>];
//# sourceMappingURL=inbox-listener.d.ts.map