"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.initializeInboxListener = void 0;
const pg_1 = require("pg");
const utils_1 = require("../common/utils");
const logical_replication_listener_1 = require("../replication/logical-replication-listener");
const concurrency_strategy_1 = require("../strategies/concurrency-strategy");
const message_processing_timeout_strategy_1 = require("../strategies/message-processing-timeout-strategy");
const message_processing_transaction_level_strategy_1 = require("../strategies/message-processing-transaction-level-strategy");
const message_retry_strategy_1 = require("../strategies/message-retry-strategy");
const poisonous_message_retry_strategy_1 = require("../strategies/poisonous-message-retry-strategy");
const inbox_message_storage_1 = require("./inbox-message-storage");
/**
 * Initialize the listener to watch for inbox table inserts.
 * @param config The configuration object with required values to connect to the WAL.
 * @param messageHandlers A list of message handlers to handle the inbox messages.
 * @param logger A logger instance for logging trace up to error logs
 * @param strategies Strategies to provide custom logic for handling specific scenarios
 * @returns Functions for a clean shutdown.
 */
const initializeInboxListener = (config, messageHandlers, logger, strategies) => {
    const pool = createPgPool(config, logger);
    const allStrategies = applyDefaultStrategies(strategies, config);
    const messageHandlerDict = getMessageHandlerDict(messageHandlers);
    const messageHandler = createMessageHandler(messageHandlerDict, pool, allStrategies, config, logger);
    const errorResolver = createErrorResolver(messageHandlerDict, pool, allStrategies, config, logger);
    const [shutdown] = (0, logical_replication_listener_1.createLogicalReplicationListener)(config, messageHandler, errorResolver, logger, allStrategies, 'inbox');
    return [
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            try {
                yield Promise.all([pool.end(), shutdown()]);
            }
            catch (e) {
                logger.error(e, 'Inbox listener shutdown error');
            }
        }),
    ];
};
exports.initializeInboxListener = initializeInboxListener;
const createPgPool = (config, logger) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        logger.error(err, 'PostgreSQL pool error');
    });
    return pool;
};
/**
 * Executes the inbox verification, the actual message handler, and marks the
 * inbox message as processed in one transaction.
 */
const createMessageHandler = (messageHandlers, pool, strategies, config, logger) => {
    return (message) => __awaiter(void 0, void 0, void 0, function* () {
        const transactionLevel = strategies.messageProcessingTransactionLevelStrategy(message);
        // increase the "started_attempts" and check if the message is probably not a poisonous message
        const attempt = yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () { return poisonousMessageCheck(message, strategies, client, config, logger); }), transactionLevel, logger);
        if (!attempt) {
            return;
        }
        yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
            const result = yield (0, inbox_message_storage_1.verifyInbox)(message, client, config);
            if (result === true &&
                strategies.messageRetryStrategy.shouldAttempt(message)) {
                const handler = messageHandlers[getMessageHandlerKey(message)];
                if (handler) {
                    yield handler.handle(message, client);
                }
                else {
                    logger.debug(`No message handler found for aggregate type "${message.aggregateType}" and message tye "${message.messageType}"`);
                }
                yield (0, inbox_message_storage_1.ackInbox)(message, client, config);
            }
            else {
                logger.warn(message, `Received inbox message cannot be processed: ${result}`);
            }
        }), transactionLevel, logger);
    });
};
/**
 * Increases the "started_attempts" counter and checks if it is now more than
 * one higher than the "finished_attempts". It calls the
 * "poisonousMessageRetryStrategy" if it is provided or checks based on the
 * configured maxPoisonousAttempts (default: 3) if another retry should be done.
 * @returns true if the message should be attempted or false if not
 */
const poisonousMessageCheck = (message, strategies, client, config, logger) => __awaiter(void 0, void 0, void 0, function* () {
    const result = yield (0, inbox_message_storage_1.poisonousMessageUpdate)(message, client, config);
    if (!result) {
        return false; // message not found
    }
    const diff = result.startedAttempts - result.finishedAttempts;
    if (diff > 1) {
        const retry = strategies.poisonousMessageRetryStrategy(message, result);
        if (!retry) {
            logger.error(message, `Stopped processing the message with ID ${message.id} as it is likely a poisonous message.`);
            yield (0, inbox_message_storage_1.ackInbox)(message, client, config);
            return false;
        }
    }
    return true;
});
const getMessageHandlerKey = (h) => `${h.aggregateType}@${h.messageType}`;
const getMessageHandlerDict = (messageHandlers) => {
    const handlers = {};
    for (const handler of messageHandlers) {
        const key = getMessageHandlerKey(handler);
        if (key in handlers) {
            throw new Error(`Only one message handler can handle one aggregate and message type. Multiple message handlers try to handle the aggregate type "${handler.aggregateType}" with the message type "${handler.messageType}".`);
        }
        handlers[key] = handler;
    }
    if (Object.keys(handlers).length === 0) {
        throw new Error('At least one message handler must be provided.');
    }
    return handlers;
};
/**
 * Increase the "finished_attempts" counter of the message and (potentially) retry it later.
 */
const createErrorResolver = (messageHandlers, pool, strategies, config, logger) => {
    /**
     * An error handler that will increase the "finished_attempts" counter on
     * transient errors or sets the "finished_attempts" counter to max attempts
     * for permanent errors. The message handler `handleError` is called to allow
     * your code to decide if the error is a transient or a permanent one.
     * @param message: the InboxMessage that failed to be processed
     * @param error: the error that was thrown while processing the message
     */
    return (message, error) => __awaiter(void 0, void 0, void 0, function* () {
        const handler = messageHandlers[getMessageHandlerKey(message)];
        const transactionLevel = strategies.messageProcessingTransactionLevelStrategy(message);
        try {
            return yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                let errorType = 'transient_error';
                const current = message.finishedAttempts + 1;
                const maxAttempts = strategies.messageRetryStrategy.maxAttempts(message);
                if (handler === null || handler === void 0 ? void 0 : handler.handleError) {
                    errorType = yield handler.handleError(error, message, client, {
                        current,
                        max: maxAttempts,
                    });
                }
                let attempts;
                if (errorType === 'permanent_error' ||
                    message.finishedAttempts + 1 >= maxAttempts) {
                    attempts = maxAttempts;
                    logger.error(error, `Giving up processing the message with id ${message.id}.`);
                }
                else {
                    logger.warn(error, `An error ocurred while processing the message with id ${message.id}.`);
                }
                yield (0, inbox_message_storage_1.nackInbox)(message, client, config, attempts);
                return errorType;
            }), transactionLevel, logger);
        }
        catch (error) {
            logger.error(Object.assign(Object.assign({}, message), { err: error }), handler
                ? 'The error handling of the message failed. Please make sure that your error handling code does not throw an error!'
                : 'The error handling of the message failed.');
            // In case the error handling logic failed do a best effort to increase the "finished_attempts" counter
            try {
                yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                    yield (0, inbox_message_storage_1.nackInbox)(message, client, config);
                }), transactionLevel, logger);
                return 'transient_error';
            }
            catch (_a) {
                return 'permanent_error';
            }
        }
    });
};
const applyDefaultStrategies = (strategies, config) => {
    var _a, _b, _c, _d, _e;
    return ({
        concurrencyStrategy: (_a = strategies === null || strategies === void 0 ? void 0 : strategies.concurrencyStrategy) !== null && _a !== void 0 ? _a : (0, concurrency_strategy_1.defaultConcurrencyStrategy)(),
        messageProcessingTimeoutStrategy: (_b = strategies === null || strategies === void 0 ? void 0 : strategies.messageProcessingTimeoutStrategy) !== null && _b !== void 0 ? _b : (0, message_processing_timeout_strategy_1.defaultMessageProcessingTimeoutStrategy)(config),
        messageProcessingTransactionLevelStrategy: (_c = strategies === null || strategies === void 0 ? void 0 : strategies.messageProcessingTransactionLevelStrategy) !== null && _c !== void 0 ? _c : (0, message_processing_transaction_level_strategy_1.defaultMessageProcessingTransactionLevelStrategy)(),
        messageRetryStrategy: (_d = strategies === null || strategies === void 0 ? void 0 : strategies.messageRetryStrategy) !== null && _d !== void 0 ? _d : (0, message_retry_strategy_1.defaultMessageRetryStrategy)(config),
        poisonousMessageRetryStrategy: (_e = strategies === null || strategies === void 0 ? void 0 : strategies.poisonousMessageRetryStrategy) !== null && _e !== void 0 ? _e : (0, poisonous_message_retry_strategy_1.defaultPoisonousMessageRetryStrategy)(config),
    });
};
//# sourceMappingURL=inbox-listener.js.map